package eu.jonahbauer.raytracing.shape;

import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.math.Vec3;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;
import java.util.Optional;

public record Sphere(@NotNull Vec3 center, double radius) implements Shape {
    public static final @NotNull Sphere UNIT = new Sphere(Vec3.ZERO, 1.0);

    public Sphere {
        Objects.requireNonNull(center, "center");
        if (radius <= 0 || !Double.isFinite(radius)) throw new IllegalArgumentException("radius must be positive");
    }

    public Sphere(double x, double y, double z, double r) {
        this(new Vec3(x, y, z), r);
    }

    @Override
    public @NotNull Optional<HitResult> hit(@NotNull Ray ray) {
        var oc = ray.origin().minus(center());

        var a = ray.direction().squared();
        var b = 2 * ray.direction().times(oc);
        var c = oc.squared() - radius * radius;

        var discriminant = b * b - 4 * a * c;
        if (discriminant < 0) return Optional.empty();

        var sd = Math.sqrt(discriminant);

        double t = (- b - sd) / (2 * a);
        if (t < 0) t = (-b + sd) / (2 * a);
        if (t < 0) return Optional.empty();
        return Optional.of(new HitResult(t, ray.at(t).minus(center)));
    }

    public @NotNull Sphere withCenter(@NotNull Vec3 center) {
        return new Sphere(center, radius);
    }

    public @NotNull Sphere withCenter(double x, double y, double z) {
        return withCenter(new Vec3(x, y, z));
    }

    public @NotNull Sphere withRadius(double radius) {
        return new Sphere(center, radius);
    }
}
